"""Tests for augur mask

NOTE: Several functions are monkeypatched in these tests. If you change the arguments
for any function in mask.py, check that it is correctly updated in this file.
"""
import argparse
import os
import pytest
import string

from Bio import SeqIO
from Bio.Seq import Seq
from Bio.SeqRecord import SeqRecord

from augur import mask
from augur.utils import VALID_NUCLEOTIDES

# Test inputs for the commands. Writing these here so tests are self-contained.
@pytest.fixture
def sequences():
    return {
        "SEQ1": SeqRecord(Seq("ATGC-ATGC-ATGC"), id="SEQ1"),
        "SEQ2": SeqRecord(Seq("ATATATATATATATAT"), id="SEQ2"),
        "SEQ_ALLCHARS": SeqRecord(Seq(string.ascii_letters + string.digits + "-"), id="SEQ_ALLCHARS")
    }

@pytest.fixture
def fasta_file(tmpdir, sequences):
    fasta_file = str(tmpdir / "test.fasta")
    with open(fasta_file, "w") as fh:
        SeqIO.write(sequences.values(), fh, "fasta")
    return fasta_file

TEST_VCF="""\
##fileformat=VCFv4.2
#CHROM	POS	ID	REF	ALT	QUAL	FILTER	INFO	FORMAT
SEQ	1	.	G	A	.		.
SEQ	2	.	G	A	.		.
SEQ	3	.	C	T	.		.
SEQ	5	.	C	T	.		.
SEQ	8	.	A	C	.		.
SEQ	13	.	A	C	.		.
"""

@pytest.fixture
def vcf_file(tmpdir):
    vcf_file = str(tmpdir / "test.vcf")
    with open(vcf_file, "w") as fh:
        fh.write(TEST_VCF)
    return vcf_file

TEST_BED_SEQUENCE = [1,4,6,7,8,9]
# IF YOU UPDATE ONE OF THESE, UPDATE THE OTHER.
TEST_BED="""\
SEQ1	1	2	IG18_Rv0018c-Rv0019c	
SEQ1	4	5	IG18_Rv0018c-Rv0019c	
SEQ1	6	8	IG18_Rv0018c-Rv0019c	
SEQ1	7	10	IG18_Rv0018c-Rv0019c	
"""

@pytest.fixture
def bed_file(tmpdir):
    bed_file = str(tmpdir / "exclude.bed")
    with open(bed_file, "w") as fh:
        fh.write(TEST_BED)
    return bed_file

TEST_MASK_SEQUENCE = [3,5,7]
TEST_MASK="""\
4
6
8
"""

@pytest.fixture
def mask_file(tmpdir):
    mask_file = str(tmpdir / "exclude.mask")
    with open(mask_file, "w") as fh:
        fh.write(TEST_MASK)
    return mask_file

@pytest.fixture
def out_file(tmpdir):
    out_file = str(tmpdir / "out")
    open(out_file, "w").close()
    return out_file

@pytest.fixture
def mp_context(monkeypatch):
    #Have found explicit monkeypatch context-ing prevents stupid bugs
    with monkeypatch.context() as mp:
        yield mp


@pytest.fixture
def argparser():
    """Provide an easy way to test command line arguments"""
    parser = argparse.ArgumentParser()
    mask.register_arguments(parser)
    def parse(args):
        return parser.parse_args(args.split(" "))
    return parse

class TestMask:
    def test_get_chrom_name_valid(self, vcf_file):
        """get_chrom_name should return the first CHROM field in a vcf file"""
        assert mask.get_chrom_name(vcf_file) == "SEQ"

    def test_get_chrom_name_invalid(self, tmpdir):
        """get_chrom_name should return nothing if no CHROM field is found in the VCF"""
        vcf_file = str(tmpdir / "incomplete.vcf")
        with open(vcf_file, "w") as fh:
            fh.write("##fileformat=VCFv4.2\n")
            fh.write("#CHROM	POS	ID	REF	ALT	QUAL	FILTER	INFO\n")
        assert mask.get_chrom_name(vcf_file) is None
    
    def test_mask_vcf_bails_on_no_chrom(self, tmpdir):
        """mask_vcf should pull a sys.exit() if get_chrom_name returns None"""
        bad_vcf = str(tmpdir / "bad.vcf")
        with open(bad_vcf, "w") as fh:
            fh.write("#")
        with pytest.raises(SystemExit) as err:
            mask.mask_vcf([], bad_vcf, "")
    
    def test_mask_vcf_creates_maskfile(self, vcf_file, mp_context):
        """mask_vcf should create a 1-indexed mask file from the 0-indexed list of sites"""
        mask_file = vcf_file + "_maskTemp"
        def shell_has_maskfile(call, **kwargs):
            assert mask_file in call
        mp_context.setattr(mask, "get_chrom_name", lambda f: "SEQ")
        mp_context.setattr(mask, "run_shell_command", shell_has_maskfile)
        mask.mask_vcf([1,5], vcf_file, vcf_file, cleanup=False)
        assert os.path.isfile(mask_file), "Mask file was not written!"
        with open(mask_file) as fh:
            assert fh.read() == "SEQ	2\nSEQ	6", "Incorrect mask file written!"
    
    def test_mask_vcf_handles_gz(self, vcf_file, mp_context):
        """mask_vcf should recognize when the in or out files are .gz and call out accordingly"""
        def test_shell(call, raise_errors=True):
            assert "--gzvcf" in call
            assert "| gzip -c" in call
        mp_context.setattr(mask, "run_shell_command", test_shell)
        mp_context.setattr(mask, "get_chrom_name", lambda f: "SEQ")
        # Still worth using the fixture here because writing the mask file uses this path
        # Using it for both entries in case god forbid someone starts assuming that path
        # valid earlier than it currently is.
        in_file = vcf_file + ".gz"
        mask.mask_vcf([1,5], in_file, in_file)

    def test_mask_vcf_removes_matching_sites(self, vcf_file, out_file):
        """mask_vcf should remove the given sites from the VCF file"""
        mask.mask_vcf([4,5], vcf_file, out_file)
        with open(out_file) as after, open(vcf_file) as before:
            assert len(after.readlines()) == len(before.readlines()) - 1, "Too many lines removed!"
            assert "SEQ\t5" not in after.read(), "Correct sites not removed!"

    def test_mask_vcf_cleanup_flag(self, vcf_file, mp_context):
        """mask_vcf should respect the cleanup flag"""
        tmp_mask_file = vcf_file + "_maskTemp"
        mp_context.setattr(mask, "run_shell_command", lambda *a, **k: None)
        mp_context.setattr(mask, "get_chrom_name", lambda f: "SEQ")

        mask.mask_vcf([], vcf_file, "", cleanup=True)
        assert not os.path.isfile(tmp_mask_file), "Temporary mask not cleaned up"

        mask.mask_vcf([], vcf_file, "", cleanup=False)
        assert os.path.isfile(tmp_mask_file), "Temporary mask cleaned up as expected"
    
    def test_mask_fasta_normal_case(self, fasta_file, out_file, sequences):
        """mask_fasta normal case - all sites in sequences"""
        mask_sites = [5,10]
        mask.mask_fasta([5,10], fasta_file, out_file)
        output = SeqIO.parse(out_file, "fasta")
        for seq in output:
            original = sequences[seq.id]
            for idx, site in enumerate(seq):
                if idx not in mask_sites:
                    assert site == original[idx], "Incorrect sites modified!"
                else:
                    assert site == "N", "Not all sites modified correctly!"
    
    def test_mask_fasta_out_of_index(self, out_file, fasta_file, sequences):
        """mask_fasta provided a list of indexes past the length of the sequences"""
        max_length = max(len(record.seq) for record in sequences.values())
        mask.mask_fasta([5, max_length, max_length+5], fasta_file, out_file)
        output = SeqIO.parse(out_file, "fasta")
        for seq in output:
            assert seq[5] == "N", "Not all sites masked correctly!"
            original = sequences[seq.id]
            for idx, site in enumerate(seq):
                if idx != 5:
                    assert site == original[idx], "Incorrect sites modified!"
    
    def test_mask_fasta_from_beginning(self, out_file, fasta_file, sequences):
        mask.mask_fasta([], fasta_file, out_file, mask_from_beginning=3)
        output = SeqIO.parse(out_file, "fasta")
        for seq in output:
            original = sequences[seq.id]
            assert seq.seq[:3] == "NNN"
            assert seq.seq[3:] == original.seq[3:]

    def test_mask_fasta_from_end(self, out_file, fasta_file, sequences):
        mask.mask_fasta([], fasta_file, out_file, mask_from_end=3)
        output = SeqIO.parse(out_file, "fasta")
        for seq in output:
            original = sequences[seq.id]
            assert seq.seq[-3:] == "NNN"
            assert seq.seq[:-3] == original.seq[:-3]

    def test_mask_fasta_from_beginning_and_end(self, out_file, fasta_file, sequences):
        mask.mask_fasta([], fasta_file, out_file, mask_from_beginning=2, mask_from_end=3)
        output = SeqIO.parse(out_file, "fasta")
        for seq in output:
            original = sequences[seq.id]
            assert seq.seq[:2] == "NN"
            assert seq.seq[-3:] == "NNN"
            assert seq.seq[2:-3] == original.seq[2:-3]

    @pytest.mark.parametrize("beginning,end", ((1000,0), (0,1000),(1000,1000)))
    def test_mask_fasta_from_beginning_and_end_too_long(self, fasta_file, out_file, beginning, end):
        mask.mask_fasta([], fasta_file, out_file, mask_from_beginning=beginning, mask_from_end=end)
        output = SeqIO.parse(out_file, "fasta")
        for record in output:
            assert record.seq == "N" * len(record.seq)

    @pytest.mark.parametrize("mask_invalid", (True, False))
    def test_mask_fasta_invalid_sites(self, fasta_file, out_file, sequences, mask_invalid):
        """Verify that mask_fasta masks invalid nucleotides when and only when mask_invalid is passed as True"""
        mask.mask_fasta([], fasta_file, out_file, mask_invalid=mask_invalid)
        output = SeqIO.to_dict(SeqIO.parse(out_file, "fasta"))
        for name, record in sequences.items():
            for site, nucleotide in enumerate(record.seq):
                if nucleotide not in VALID_NUCLEOTIDES and mask_invalid is True:
                    assert output[name][site] == "N"
                else:
                    assert output[name][site] == nucleotide

    def test_run_handle_missing_sequence_file(self, vcf_file, argparser):
        os.remove(vcf_file)
        args = argparser("-s %s" % vcf_file)
        with pytest.raises(SystemExit):
            mask.run(args)

    def test_run_handle_empty_sequence_file(self, vcf_file, argparser):
        open(vcf_file,"w").close()
        args = argparser("-s %s --mask-sites 1" % vcf_file)
        with pytest.raises(SystemExit):
            mask.run(args)

    def test_run_handle_missing_mask_file(self, vcf_file, bed_file, argparser):
        os.remove(bed_file)
        args = argparser("-s %s --mask %s" % (vcf_file, bed_file))
        with pytest.raises(SystemExit):
            mask.run(args)

    def test_run_handle_empty_mask_file(self, vcf_file, bed_file, argparser):
        open(bed_file, "w").close()
        args = argparser("-s %s --mask %s" % (vcf_file, bed_file))
        with pytest.raises(SystemExit):
            mask.run(args)

    def test_run_recognize_vcf(self, bed_file, vcf_file, argparser, mp_context):
        """Ensure we're handling vcf files correctly"""
        args = argparser("--mask=%s -s %s --no-cleanup" % (bed_file, vcf_file))
        def fail(*args, **kwargs):
            assert False, "Called mask_fasta incorrectly"
        mp_context.setattr(mask, "mask_vcf", lambda *a, **k: None)
        mp_context.setattr(mask, "mask_fasta", fail)
        mp_context.setattr(mask, "copyfile", lambda *args: None)
        mask.run(args)

    def test_run_recognize_fasta(self, bed_file, fasta_file, argparser, mp_context):
        """Ensure we're handling fasta files correctly"""
        args = argparser("--mask=%s -s %s --no-cleanup" % (bed_file, fasta_file))
        def fail(*args, **kwargs):
            assert False, "Called mask_fasta incorrectly"
        mp_context.setattr(mask, "mask_fasta", lambda *a, **k: None)
        mp_context.setattr(mask, "mask_vcf", fail)
        mp_context.setattr(mask, "copyfile", lambda *args: None)
        mask.run(args)

    def test_run_handle_missing_outfile(self, bed_file, fasta_file, argparser, mp_context):
        args = argparser("--mask=%s -s %s" % (bed_file, fasta_file))
        expected_outfile = os.path.join(os.path.dirname(fasta_file), "masked_" + os.path.basename(fasta_file))
        def check_outfile(mask_sites, in_file, out_file, **kwargs):
            assert out_file == expected_outfile
            with open(out_file, "w") as fh:
                fh.write("test_string")
        mp_context.setattr(mask, "mask_fasta", check_outfile)
        mask.run(args)
        with open(fasta_file) as fh:
            assert fh.read() == "test_string"
    
    def test_run_respect_no_cleanup(self, bed_file, vcf_file, argparser, mp_context):
        out_file = os.path.join(os.path.dirname(vcf_file), "masked_" + os.path.basename(vcf_file))
        def make_outfile(mask_sites, in_file, out_file, cleanup=True):
            assert cleanup == False
            open(out_file, "w").close() # need out_file to exist
        mp_context.setattr(mask, "mask_vcf", make_outfile)
        args = argparser("--mask=%s -s %s -o %s --no-cleanup" % (bed_file, vcf_file, out_file))
        mask.run(args)
        assert os.path.exists(out_file), "Output file incorrectly deleted"

    def test_run_normal_case(self, bed_file, vcf_file, out_file, argparser, mp_context):
        def check_args(mask_sites, in_file, _out_file, cleanup):
            assert mask_sites == TEST_BED_SEQUENCE, "Wrong mask sites provided"
            assert in_file == vcf_file, "Incorrect input file provided"
            assert _out_file == out_file, "Incorrect output file provided"
            assert cleanup is True, "Cleanup erroneously passed in as False"
        mp_context.setattr(mask, "mask_vcf", check_args)
        args = argparser("--mask=%s --sequences=%s --output=%s" %(bed_file, vcf_file, out_file))
        mask.run(args)
        assert os.path.exists(out_file), "Output file incorrectly deleted"
    
    def test_run_with_mask_sites(self, vcf_file, out_file, argparser, mp_context):
        args = argparser("--mask-sites 2 8 -s %s -o %s" % (vcf_file, out_file))
        def check_mask_sites(mask_sites, *args, **kwargs):
            # mask-sites are passed to the CLI as one-indexed
            assert mask_sites == [1,7]
        mp_context.setattr(mask, "mask_vcf", check_mask_sites)
        mask.run(args)

    def test_run_with_mask_sites_and_mask_file(self, vcf_file, out_file, bed_file, argparser, mp_context):
        args = argparser("--mask-sites 20 21 --mask %s -s %s -o %s" % (bed_file, vcf_file, out_file))
        def check_mask_sites(mask_sites, *args, **kwargs):
            # mask-sites are passed to the CLI as one-indexed
            assert mask_sites == sorted(set(TEST_BED_SEQUENCE + [19,20]))
        mp_context.setattr(mask, "mask_vcf", check_mask_sites)
        mask.run(args)

    def test_run_requires_some_masking(self, vcf_file, argparser):
        args = argparser("-s %s" % vcf_file)
        with pytest.raises(SystemExit) as err:
            mask.run(args)

    @pytest.mark.parametrize("op", ("beginning", "end"))
    def test_run_vcf_cannot_mask_beginning_or_end(self, vcf_file, argparser, op):
        args = argparser("-s %s --mask-from-%s 2" % (vcf_file, op))
        with pytest.raises(SystemExit) as err:
            mask.run(args)

    def test_run_fasta_mask_from_beginning_or_end(self, fasta_file, out_file, argparser, mp_context):
        args = argparser("-s %s -o %s --mask-from-beginning 2 --mask-from-end 3" % (fasta_file, out_file))
        def check_mask_from(*args, mask_from_beginning=0, mask_from_end=0, **kwargs):
            assert mask_from_beginning == 2
            assert mask_from_end == 3
        mp_context.setattr(mask, "mask_fasta", check_mask_from)
        mask.run(args)

    def test_e2e_fasta_minimal(self, fasta_file, bed_file, sequences, argparser):
        args = argparser("-s %s --mask %s" % (fasta_file, bed_file))
        mask.run(args)
        output = SeqIO.parse(fasta_file,"fasta")
        for record in output:
            reference = sequences[record.id].seq
            for idx, site in enumerate(record.seq):
                if idx in TEST_BED_SEQUENCE:
                    assert site == "N"
                else:
                    assert site == reference[idx]

    def test_e2e_fasta_mask_file(self, fasta_file, mask_file, sequences, argparser):
        args = argparser("-s %s --mask %s" % (fasta_file, mask_file))
        mask.run(args)
        output = SeqIO.parse(fasta_file,"fasta")
        for record in output:
            reference = sequences[record.id].seq
            for idx, site in enumerate(record.seq):
                if idx in TEST_MASK_SEQUENCE:
                    assert site == "N"
                else:
                    assert site == reference[idx]

    def test_e2e_fasta_beginning_end_sites(self, fasta_file, bed_file, out_file, sequences, argparser):
        from_beginning = 3
        from_end = 1
        arg_sites = [5, 12]
        expected_removals = sorted(set(TEST_BED_SEQUENCE + [s - 1 for s in arg_sites]))
        print(expected_removals)
        args = argparser("-s %s -o %s --mask %s --mask-from-beginning %s --mask-from-end %s --mask-sites %s" % (
                         fasta_file, out_file, bed_file, from_beginning, from_end, " ".join(str(s) for s in arg_sites)))
        mask.run(args)
        output = SeqIO.parse(out_file, "fasta")
        for record in output:
            reference = str(sequences[record.id].seq)
            masked_seq = str(record.seq)
            assert masked_seq[:from_beginning] == "N" * from_beginning
            assert masked_seq[-from_end:] == "N" * from_end
            for idx, site in enumerate(masked_seq[from_beginning:-from_end], from_beginning):
                if idx in expected_removals:
                    assert site == "N"
                else:
                    assert site == reference[idx]

    def test_e2e_fasta_mask_invalid(self, fasta_file, out_file, sequences, argparser):
        args = argparser("-s %s -o %s --mask-invalid" % (fasta_file, out_file))
        mask.run(args)
        output = SeqIO.parse(out_file, "fasta")
        for record in output:
            reference = str(sequences[record.id].seq)
            for idx, site in enumerate(reference):
                assert record.seq[idx] == site if site in VALID_NUCLEOTIDES else "N"

    def test_e2e_vcf_minimal(self, vcf_file, bed_file, argparser):
        args = argparser("-s %s --mask %s" % (vcf_file, bed_file))
        mask.run(args)
        with open(vcf_file) as output:
            assert output.readline().startswith("##fileformat") # is a VCF
            assert output.readline().startswith("#CHROM\tPOS\t") # have a header
            for line in output.readlines():
                site = int(line.split("\t")[1]) # POS column
                site = site - 1 # shift to zero-indexed site
                assert site not in TEST_BED_SEQUENCE

    def test_e2e_vcf_mask_file(self, vcf_file, mask_file, argparser):
        args = argparser("-s %s --mask %s" % (vcf_file, mask_file))
        mask.run(args)
        with open(vcf_file) as output:
            assert output.readline().startswith("##fileformat") # is a VCF
            assert output.readline().startswith("#CHROM\tPOS\t") # have a header
            for line in output.readlines():
                site = int(line.split("\t")[1]) # POS column
                site = site - 1 # shift to zero-indexed site
                assert site not in TEST_MASK_SEQUENCE

    def test_e2e_vcf_with_options(self, vcf_file, bed_file, out_file, argparser):
        arg_sites = [5, 12, 14]
        expected_removals = sorted(set(TEST_BED_SEQUENCE + [s - 1 for s in arg_sites]))
        args = argparser("-s %s -o %s --mask %s --mask-sites %s" % (
                         vcf_file, out_file, bed_file, " ".join(str(s) for s in arg_sites)))
        mask.run(args)
        with open(out_file) as output:
            assert output.readline().startswith("##fileformat") # is a VCF
            assert output.readline().startswith("#CHROM\tPOS\t") # have a header
            for line in output.readlines():
                site = int(line.split("\t")[1]) # POS column
                site = site - 1 #re-zero-index the VCF sites
                assert site not in expected_removals



from itertools import product

year = {'%y', '%Y'}
month = {'%b', '%B', '%m'}
day = {'%d'}
month_and_day = {'%j'}
week = {'%U', '%W'}
day_of_week = {'%A', '%a', '%w', '%u'}

# Set of directives that can be converted to complete date with year, month, and day
YEAR_MONTH_DAY_DIRECTIVES = (
    # Locale's full date representation
    {('%c',),('%x',)} |
    # Dates with ISO 8601 week dates for year ('%G' is NOT interchangeable with '%Y'), ISO 8601 week ('%V'), and weekdays
    {('%G', '%V', '%A'),('%G', '%V', '%a'),('%G', '%V', '%w'),('%G', '%V', '%u')} |
    # Dates with year, week, and weekday
    set(product(year, week, day_of_week)) |
    # Dates with year and day of the year
    set(product(year, month_and_day)) |
    # Dates with year, month, and day
    set(product(year, month, day))
)

# Set of directives that can be converted to incomplete dates, missing the day
YEAR_MONTH_DIRECTIVES = set(product(year, month))

# Set of directives that can be converted to incomplete dates, missing the month and day
YEAR_DIRECTIVES = set(product(year))



"""
Infer ancestral traits based on a tree.
"""

import numpy as np
from collections import defaultdict
import sys
from .argparse_ import ExtendOverwriteDefault
from .errors import AugurError
from .io.file import open_file
from .io.metadata import DEFAULT_DELIMITERS, DEFAULT_ID_COLUMNS, InvalidDelimiter, read_metadata
from .utils import write_json, get_json_name
TINY = 1e-12

def mugration_inference(tree=None, seq_meta=None, field='country', confidence=True,
                        missing='?', sampling_bias_correction=None, weights=None):
    """
    Infer likely ancestral states of a discrete character assuming a time reversible model.

    Parameters
    ----------
    tree : str
        name of tree file
    seq_meta : pandas.DataFrame
        meta data associated with sequences
    field : str, optional
        meta data field to use
    confidence : bool, optional
        calculate confidence values for inferences
    missing : str, optional
        character that is to be interpreted as missing data, default='?'
    sampling_bias_correction : None, optional
        factor by which the transition rate is scaled up to counter sampling bias
    weights : None, optional
        vector of equilibrium frequencies that one expects the far ancestor to be sampled from

    Returns
    -------
    T : Bio.Phylo.BaseTree.Tree
        Biophyton tree
    gtr : treetime.GTR
        GTR model
    alphabet : dict
        mapping of character states to
    """
    from treetime.wrappers import reconstruct_discrete_traits
    from Bio import Phylo

    T = Phylo.read(tree, 'newick')
    traits = {}
    nodes = {n.name:n for n in T.get_terminals()}
    for name, meta in seq_meta.iterrows():
        if field in meta and name in nodes and meta[field] != missing:
            traits[name] = meta[field]
    unique_states = list(set(traits.values()))

    if len(unique_states)==0:
        print("WARNING: no states found for discrete state reconstruction.", file=sys.stderr)
        for node in T.find_clades():
            node.__setattr__(field, None)
        return T, None, {}
    elif len(unique_states)==1:
        print("WARNING: only one state found for discrete state reconstruction:", unique_states, file=sys.stderr)
        for node in T.find_clades():
            node.__setattr__(field, unique_states[0])
        return T, None, {}
    elif len(unique_states)<300:
        tt, letter_to_state, reverse_alphabet = \
            reconstruct_discrete_traits(T, traits, missing_data=missing,
                 sampling_bias_correction=sampling_bias_correction, weights=weights)
    else:
        print("ERROR: 300 or more distinct discrete states found. TreeTime is currently not set up to handle that many states.", file=sys.stderr)
        sys.exit(1)

    if tt is None:
        print("ERROR in discrete state reconstruction in TreeTime. Please look for errors above.", file=sys.stderr)
        sys.exit(1)

    # attach inferred states as e.g. node.region = 'africa'
    for node in tt.tree.find_clades():
        node.__setattr__(field, letter_to_state[node.cseq[0]])

    # if desired, attach entropy and confidence as e.g. node.region_entropy = 0.03
    if confidence:
        for node in tt.tree.find_clades():
            pdis = node.marginal_profile[0]
            S = -np.sum(pdis*np.log(pdis+TINY))

            marginal = [(letter_to_state[tt.gtr.alphabet[i]], pdis[i]) for i in range(len(tt.gtr.alphabet))]
            marginal.sort(key=lambda x: x[1], reverse=True) # sort on likelihoods
            # Values are defined for all demes, although many/most will be 0, so only take those over 0.1%
            marginal = [(a, b) for a, b in marginal if b > 0.001]
            conf = {a:b for a,b in marginal}
            node.__setattr__(field + "_entropy", S)
            node.__setattr__(field + "_confidence", conf)

    return tt.tree, tt.gtr, letter_to_state


def register_parser(parent_subparsers):
    parser = parent_subparsers.add_parser("traits", help=__doc__)
    parser.add_argument('--tree', '-t', required=True, help="tree to perform trait reconstruction on")
    parser.add_argument('--metadata', required=True, metavar="FILE", help="table with metadata")
    parser.add_argument('--metadata-delimiters', default=DEFAULT_DELIMITERS, nargs="+", action=ExtendOverwriteDefault,
                        help="delimiters to accept when reading a metadata file. Only one delimiter will be inferred.")
    parser.add_argument('--metadata-id-columns', default=DEFAULT_ID_COLUMNS, nargs="+", action=ExtendOverwriteDefault,
                        help="names of possible metadata columns containing identifier information, ordered by priority. Only one ID column will be inferred.")
    parser.add_argument('--weights', required=False, help="tsv/csv table with equilibrium probabilities of discrete states")
    parser.add_argument('--columns', required=True, nargs='+', action='extend',
                        help='metadata fields to perform discrete reconstruction on')
    parser.add_argument('--confidence',action="store_true",
                        help='record the distribution of subleading mugration states')
    parser.add_argument('--sampling-bias-correction', type=float,
                        help='a rough estimate of how many more events would have been observed'
                             ' if sequences represented an even sample. This should be'
                             ' roughly the (1-sum_i p_i^2)/(1-sum_i t_i^2), where p_i'
                             ' are the equilibrium frequencies and t_i are apparent ones.'
                             '(or rather the time spent in a particular state on the tree)')
    parser.add_argument('--output-node-data', type=str, help='name of JSON file to save trait inferences to')
    parser.epilog = "Note that missing data must be represented by a `?` character. Missing data will currently be inferred."
    return parser


def run(args):
    """run mugration inference

    Parameters
    ----------
    args : argparse.Namespace
        command line arguments are parsed by argparse
    """
    tree_fname = args.tree
    try:
        traits = read_metadata(
            args.metadata,
            delimiters=args.metadata_delimiters,
            id_columns=args.metadata_id_columns,

            # Read all columns as string for discrete trait analysis
            dtype="string",
        )
    except InvalidDelimiter:
        raise AugurError(
                f"Could not determine the delimiter of {args.metadata!r}. "
                f"Valid delimiters are: {args.metadata_delimiters!r}. "
                "This can be changed with --metadata-delimiters."
            )

    from Bio import Phylo
    T = Phylo.read(tree_fname, 'newick')
    missing_internal_node_names = [n.name is None for n in T.get_nonterminals()]
    if np.all(missing_internal_node_names):
        print("\n*** WARNING: Tree has no internal node names!", file=sys.stderr)
        print("*** Without internal node names, ancestral traits can't be linked up to the correct node later.", file=sys.stderr)
        print("*** If you want to use 'augur export' later, re-run this command with the output of 'augur refine'.", file=sys.stderr)
        print("*** If you haven't run 'augur refine', you can add node names to your tree by running:", file=sys.stderr)
        print("*** augur refine --tree %s --output-tree <filename>.nwk" % (tree_fname), file=sys.stderr)
        print("*** And use <filename>.nwk as the tree when running 'ancestral', 'translate', and 'traits'", file=sys.stderr)

    if args.weights:
        weight_dict = {c:{} for c in args.columns}
        sep = ',' if args.weights.endswith('csv') else '\t'
        with open_file(args.weights, 'r') as fh:
            for line in fh:
                if line[0]=='#':
                    continue
                name, trait, value = line.strip().split(sep)
                if name in weight_dict:
                    weight_dict[name][trait] = float(value)
        for c in weight_dict:
            if len(weight_dict[c])==0:
                weight_dict[c]=None
    else:
        weight_dict = {c:None for c in args.columns}

    mugration_states = defaultdict(dict)
    models = defaultdict(dict)
    out_prefix = '.'.join(args.output_node_data.split('.')[:-1])

    from treetime import version as treetime_version
    print(f"augur traits is using TreeTime version {treetime_version}")

    for column in args.columns:
        T, gtr, alphabet = mugration_inference(tree=tree_fname, seq_meta=traits,
                                               field=column, confidence=args.confidence,
                                               sampling_bias_correction=args.sampling_bias_correction,
                                               weights=weight_dict[column])
        if T is None: # something went wrong
            continue

        for node in T.find_clades():
            mugration_states[node.name][column] = getattr(node, column)

            if args.confidence:
                confidence = getattr(node, f"{column}_confidence", None)
                if confidence is not None:
                    mugration_states[node.name][f"{column}_confidence"] = confidence

                entropy = getattr(node, f"{column}_entropy", None)
                if entropy is not None:
                    mugration_states[node.name][f"{column}_entropy"] = entropy

        if gtr:
            # add gtr models to json structure for export
            models[column]['rate'] = gtr.mu
            models[column]['alphabet'] = [alphabet[k] for k in sorted(alphabet.keys())]
            models[column]['equilibrium_probabilities'] = list(gtr.Pi)
            models[column]['transition_matrix'] = [list(x) for x in gtr.W]

        if gtr:
            with open_file(out_prefix+'%s.mugration_model.txt'%column, 'w') as ofile:
                ofile.write('Map from character to field name\n')
                for k,v in alphabet.items():
                    ofile.write(k+':\t'+str(v)+'\n')
                ofile.write('\n\n')

                ofile.write(str(gtr))

    out_name = get_json_name(args, out_prefix+'_traits.json')
    write_json({"models":models, "nodes":mugration_states},out_name)

    print("\nInferred ancestral states of discrete character using TreeTime:"
          "\n\tSagulenko et al. TreeTime: Maximum-likelihood phylodynamic analysis"
          "\n\tVirus Evolution, vol 4, https://academic.oup.com/ve/article/4/1/vex042/4794731\n", file=sys.stdout)

    print("results written to", out_name, file=sys.stdout)



from pathlib    import Path
import setuptools
import sys

py_min_version = (3, 8)  # minimal supported python version
since_augur_version = (23, 0)  # py_min_version is required since this augur version

if sys.version_info < py_min_version:
    error = """
Beginning with augur {0}, Python {1} or above is required.
You are using Python {2}.

This may be due to an out of date pip.

Make sure you have pip >= 9.0.1.
""".format('.'.join(str(n) for n in since_augur_version),
           '.'.join(str(n) for n in py_min_version),
           '.'.join(str(n) for n in sys.version_info[:3]))
    sys.exit(error)

base_dir = Path(__file__).parent.resolve()
version_file = base_dir / "augur/__version__.py"
readme_file = base_dir / "README.md"

# Eval the version file to get __version__; avoids importing our own package
with version_file.open() as f:
    exec(f.read())

# Get the long description from the README file
with readme_file.open(encoding = "utf-8") as f:
    long_description = f.read()



setuptools.setup(
    name = "nextstrain-augur",
    version = __version__,  # noqa: F821; This is imported from version_file.
    author = "Nextstrain developers",
    author_email = "trevor@bedford.io, richard.neher@unibas.ch",
    description = "A bioinformatics toolkit for phylogenetic analysis",
    long_description = long_description,
    long_description_content_type = "text/markdown",
    keywords = "nextstrain, molecular epidemiology",
    url = "https://github.com/nextstrain/augur",
    project_urls = {
        "Bug Reports": "https://github.com/nextstrain/augur/issues",
        "Change Log": "https://github.com/nextstrain/augur/blob/-/CHANGES.md#next",
        "Source": "https://github.com/nextstrain/augur",
    },
    packages = setuptools.find_packages(),
    package_data = {'augur': ['data/*']},
    python_requires = '>={}'.format('.'.join(str(n) for n in py_min_version)),
    install_requires = [
        "bcbio-gff >=0.7.1, ==0.7.*",
        # TODO: Remove biopython >= 1.80 pin if it is added to bcbio-gff: https://github.com/chapmanb/bcbb/issues/142
        "biopython >=1.80, ==1.*",
        "cvxopt >=1.1.9, ==1.*",
        "importlib_resources >=5.3.0; python_version < '3.11'",
        "isodate ==0.6.*",
        "jsonschema >=3.0.0, ==3.*",
        "networkx >= 2.5, <4",
        "numpy ==1.*",
        "packaging >=19.2",
        "pandas >=1.0.0, <3",
        "phylo-treetime >=0.11.2, <0.12",
        "pyfastx >=1.0.0, <3.0",
        "python_calamine >=0.2.0",
        "scipy ==1.*",
        "xopen[zstd] >=1.7.0, <3" # TODO: Deprecated, remove v1 support around November 2024
    ],
    extras_require = {
        'dev': [
            "cram >=0.7",
            "deepdiff >=4.3.2, <8.0.0",
            "flake8",
            "freezegun >=0.3.15",
            "mypy",
            "nextstrain-sphinx-theme >=2022.5",
            "pandas-stubs >=1.0.0, <3",
            "pylint >=1.7.6",
            "pytest >=5.4.1",
            "pytest-cov >=2.8.1",
            "pytest-mock >= 2.0.0",
            "recommonmark >=0.5.0",
            "Sphinx >=2.0.1",
            "sphinx-autobuild >=2021.3.14",
            "sphinx-argparse >=0.2.5, !=0.5.0",
            "sphinx-markdown-tables >= 0.0.9",
            "sphinx-rtd-theme >=0.4.3",
            "sphinx-autodoc-typehints >=1.21.4",
            "sphinx-tabs",
            "types-jsonschema >=3.0.0, ==3.*",
            "types-setuptools",
            "wheel >=0.32.3",
            "ipdb >=0.10.1"
        ]
    },
    classifiers = [
        "Development Status :: 5 - Production/Stable",
        "Topic :: Scientific/Engineering :: Bio-Informatics",
        "License :: OSI Approved :: GNU Affero General Public License v3",

        # Python 3 only
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
    ],
    # Install an "augur" program which calls augur.__main__.main()
    #   https://setuptools.readthedocs.io/en/latest/setuptools.html#automatic-script-creation
    entry_points = {
        "console_scripts": [
            "augur = augur.__main__:main",
        ]
    }
)



"""
Mask specified sites from a VCF or FASTA file.
"""
import os
import sys
from shutil import copyfile

from Bio.Seq import MutableSeq

from .io.file import open_file
from .io.sequences import read_sequences, write_sequences
from .io.shell_command_runner import run_shell_command
from .io.vcf import shquote, is_vcf
from .utils import load_mask_sites, VALID_NUCLEOTIDES

def get_chrom_name(vcf_file):
    """Read the CHROM field from the first non-header line of a vcf file.

    Returns:
    str or None: Either the CHROM field or None if no non-comment line could be found.
    """
    with open_file(vcf_file, mode='r') as f:
        for line in f:
            if line[0] != "#":
                header = line.strip().partition('\t')
                return header[0]

def mask_vcf(mask_sites, in_file, out_file, cleanup=True):
    """Mask the provided site list from a VCF file and write to a new file.

    This function relies on 'vcftools --exclude-positions' to mask the requested sites.

    Parameters
    ----------
    mask_sites: list of int
        A list of site indexes to exclude from the vcf.
    in_file: str
        The path to the vcf file you wish to mask.
    out_file: str
        The path to write the resulting vcf to
    cleanup: bool
        Clean up the intermediate files, including the VCFTools log and mask sites file
    """
    cleanup_files = ['out.log']

    # Create the temporary masking file to be passed to VCFTools
    # Need CHROM name from VCF file:
    chrom_name = get_chrom_name(in_file)
    if chrom_name is None:
        print("ERROR: Something went wrong reading your VCF file: a CHROM column could not be found. "
              "Please check the file is valid VCF format.")
        sys.exit(1)

    # mask_sites is zero-indexed, VCFTools expects 1-indexed.
    exclude = [chrom_name + "\t" + str(pos + 1) for pos in mask_sites]
    temp_mask_file = in_file + "_maskTemp"
    with open_file(temp_mask_file, 'w') as fh:
        fh.write("\n".join(exclude))
    cleanup_files.append(temp_mask_file)

    #Read in/write out according to file ending
    in_call = "--gzvcf" if in_file.lower().endswith(".gz") else "--vcf"
    out_call = "| gzip -c" if out_file.lower().endswith(".gz") else ""

    call = ["vcftools", "--exclude-positions", shquote(temp_mask_file), in_call, shquote(in_file), "--recode --stdout", out_call, ">", shquote(out_file)]
    print("Removing masked sites from VCF file using vcftools... this may take some time. Call:")
    print(" ".join(call))
    run_shell_command(" ".join(call), raise_errors = True)
    # remove vcftools log file
    if cleanup:
        for file in cleanup_files:
            try:
                os.remove(file)
            except OSError:
                pass


def mask_sequence(sequence, mask_sites, mask_from_beginning, mask_from_end, mask_invalid):
    """Mask characters at the given sites in a single sequence record, modifying the
    record in place.

    Parameters
    ----------
    sequence : Bio.SeqRecord.SeqRecord
        A sequence to be masked
    mask_sites: list of int
        A list of site indexes to exclude from the FASTA.
    mask_from_beginning: int
        Number of sites to mask from the beginning of each sequence (default 0)
    mask_from_end: int
        Number of sites to mask from the end of each sequence (default 0)
    mask_invalid: bool
        Mask invalid nucleotides (default False)

    Returns
    -------
    Bio.SeqRecord.SeqRecord
        Masked sequence in its original record object

    """
    # Convert to a mutable sequence to enable masking with Ns.
    sequence_length = len(sequence.seq)
    beginning, end = mask_from_beginning, mask_from_end

    if beginning + end > sequence_length:
        beginning, end = sequence_length, 0

    seq = str(sequence.seq)[beginning:-end or None]

    if mask_invalid:
        seq = "".join(nuc if nuc in VALID_NUCLEOTIDES else "N" for nuc in seq)

    masked_sequence = MutableSeq("N" * beginning + seq + "N" * end)

    # Replace all excluded sites with Ns.
    for site in mask_sites:
        if site < sequence_length:
            masked_sequence[site] = "N"

    sequence.seq = masked_sequence

    return sequence


def mask_fasta(mask_sites, in_file, out_file, mask_from_beginning=0, mask_from_end=0, mask_invalid=False):
    """Mask the provided site list from a FASTA file and write to a new file.

    Masked sites are overwritten as "N"s.

    Parameters
    ----------
    mask_sites: list of int
        A list of site indexes to exclude from the FASTA.
    in_file: str
        The path to the FASTA file you wish to mask.
    out_file: str
        The path to write the resulting FASTA to
    mask_from_beginning: int
       Number of sites to mask from the beginning of each sequence (default 0)
    mask_from_end: int
       Number of sites to mask from the end of each sequence (default 0)
    mask_invalid: bool
        Mask invalid nucleotides (default False)
    """
    # Load alignment as FASTA generator to prevent loading the whole alignment
    # into memory.
    alignment = read_sequences(in_file)

    # Write the masked alignment to disk one record at a time.
    print("Removing masked sites from FASTA file.")

    masked_sequences = (
        mask_sequence(
            sequence,
            mask_sites,
            mask_from_beginning,
            mask_from_end,
            mask_invalid,
        )
        for sequence in alignment
    )
    sequences_written = write_sequences(
        masked_sequences,
        out_file,
        "fasta"
    )

def register_arguments(parser):
    """
    Add arguments to parser.
    Kept as a separate function than `register_parser` to continue to support
    unit tests that use this function to create argparser.
    """
    parser.add_argument('--sequences', '-s', required=True, help="sequences in VCF or FASTA format")
    parser.add_argument('--mask', dest="mask_file", required=False, help="locations to be masked in either BED file format, DRM format, or one 1-indexed site per line.")
    parser.add_argument('--mask-from-beginning', type=int, default=0, help="FASTA Only: Number of sites to mask from beginning")
    parser.add_argument('--mask-from-end', type=int, default=0, help="FASTA Only: Number of sites to mask from end")
    parser.add_argument('--mask-invalid', action='store_true', help="FASTA Only: Mask invalid nucleotides")
    parser.add_argument("--mask-sites", nargs='+', action='extend', type = int,  help="1-indexed list of sites to mask")
    parser.add_argument('--output', '-o', help="output file")
    parser.add_argument('--no-cleanup', dest="cleanup", action="store_false",
                        help="Leave intermediate files around. May be useful for debugging")

def register_parser(parent_subparsers):
    parser = parent_subparsers.add_parser("mask", help=__doc__)
    register_arguments(parser)
    return parser


def run(args):
    '''
    Mask specified sites from the VCF or FASTA.

    For VCF files, his occurs by removing them entirely from the VCF, essentially making
    them identical to the reference at the locations.

    For FASTA files, masked sites are replaced with "N".

    If users don't specify output, will overwrite the input file.
    '''
    # Check files exist and are not empty
    if not os.path.isfile(args.sequences):
        print("ERROR: File {} does not exist!".format(args.sequences))
        sys.exit(1)
    if os.path.getsize(args.sequences) == 0:
        print("ERROR: {} is empty. Please check how this file was produced. "
              "Did an error occur in an earlier step?".format(args.sequences))
        sys.exit(1)
    if args.mask_file:
        if not os.path.isfile(args.mask_file):
            print("ERROR: File {} does not exist!".format(args.mask_file))
            sys.exit(1)
        if os.path.getsize(args.mask_file) == 0:
            print("ERROR: {} is an empty file.".format(args.mask_file))
            sys.exit(1)
    if not any((args.mask_file, args.mask_from_beginning, args.mask_from_end, args.mask_sites, args.mask_invalid)):
        print("No masking sites provided. Must include one of --mask, --mask-from-beginning, --mask-from-end, --mask-invalid, or --mask-sites")
        sys.exit(1)

    mask_sites = set()
    if args.mask_sites:
        # Mask sites passed in as 1-indexed
        mask_sites.update(site - 1 for site in args.mask_sites)
    if args.mask_file:
        mask_sites.update(load_mask_sites(args.mask_file))
    mask_sites = sorted(mask_sites)

    # For both FASTA and VCF masking, we need a proper separate output file
    if args.output is not None:
        out_file = args.output
    else:
        out_file = os.path.join(os.path.dirname(args.sequences),
                                "masked_" + os.path.basename(args.sequences))

    if is_vcf(args.sequences):
        if args.mask_from_beginning or args.mask_from_end or args.mask_invalid:
            print("Cannot use --mask-from-beginning, --mask-from-end, or --mask-invalid with VCF files!")
            sys.exit(1)
        mask_vcf(mask_sites, args.sequences, out_file, args.cleanup)
    else:
        mask_fasta(mask_sites, args.sequences, out_file,
                   mask_from_beginning=args.mask_from_beginning,
                   mask_from_end=args.mask_from_end,
                   mask_invalid=args.mask_invalid)

    if args.output is None:
        copyfile(out_file, args.sequences)
        if args.cleanup:
            os.remove(out_file)



"""
Translate gene regions from nucleotides to amino acids.

Translates nucleotide sequences of nodes in a tree to amino acids for gene
regions of the annotated features of the provided reference sequence.
Each node then gets assigned a list of amino acid mutations for any position
that has a mismatch between its own amino acid sequence and its parent's sequence.
The reference amino acid sequences, genome annotations, and node amino acid
mutations are output to a node-data JSON file.

.. note::

    The mutation positions in the node-data JSON are one-based.
"""

import sys
import numpy as np
from Bio import SeqIO, Seq, SeqRecord, Phylo
from .io.vcf import write_VCF_translation, is_vcf as is_filename_vcf
from .utils import parse_genes_argument, read_node_data, load_features, \
    write_json, get_json_name, genome_features_to_auspice_annotation
from treetime.vcf_utils import read_vcf
from augur.errors import AugurError
from textwrap import dedent
from .argparse_ import add_validation_arguments
from .util_support.node_data_file import NodeDataObject

class MissingNodeError(Exception):
    pass

class MismatchNodeError(Exception):
    pass

class NoVariationError(Exception):
    pass

def safe_translate(sequence, report_exceptions=False):
    """Returns an amino acid translation of the given nucleotide sequence accounting
    for gaps in the given sequence.

    Optionally, returns a tuple of the translated sequence and whether an
    exception was raised during initial translation.

    Examples
    --------
    >>> safe_translate("ATG")
    'M'
    >>> safe_translate("ATGGT-")
    'MX'
    >>> safe_translate("ATG---")
    'M-'
    >>> safe_translate("ATGTAG")
    'M*'
    >>> safe_translate("")
    ''
    >>> safe_translate("ATGT")
    'MX'
    >>> safe_translate("ATG", report_exceptions=True)
    ('M', False)
    >>> safe_translate("ATGA-G", report_exceptions=True)
    ('MX', True)
    """
    from Bio.Data.CodonTable import TranslationError
    from Bio.Seq import CodonTable
    translation_exception = False

    #sequences not mod 3 give messy BiopythonWarning, so avoid by padding.
    if len(sequence)%3:
        sequence_padded = sequence + "N"*(3-len(sequence)%3)
    else:
        sequence_padded = sequence
    try:
        # Attempt translation by extracting the sequence according to the
        # BioPhython SeqFeature in frame gaps of three will translate as '-'
        translated_sequence = str(Seq.Seq(sequence_padded).translate(gap='-'))
    except TranslationError:
        translation_exception = True
        # Any other codon like '-AA' or 'NNT' etc will fail. Translate codons
        # one by one.
        codon_table  = CodonTable.ambiguous_dna_by_name['Standard'].forward_table
        str_seq = str(sequence_padded)
        codons = np.frombuffer(str_seq[:len(str_seq) - len(str_seq) % 3].encode(), dtype='S3').astype("U")
        assert len(codons) > 0
        aas = []

        for c in codons:
            # Parse result of single codon translation, add amino acids as
            # appropriate.
            try:
                aa = codon_table.get(c)
                if aa is None:
                    if c == '---':
                        aas.append('-')
                    else:
                        aas.append('X')
                else:
                    aas.append(aa)
            except (TranslationError, ValueError):
                aas.append('X')

        translated_sequence = "".join(aas)

    if report_exceptions:
        return translated_sequence, translation_exception
    else:
        return translated_sequence


def translate_feature(aln, feature):
    '''
    Translates a subsequence of input nucleotide sequences.

    Parameters
    ----------
    aln : dict
        sequences indexed by node name

    feature : Bio.Seq.Seq
        BioPython sequence feature

    Returns
    -------
    dict :
        translated sequences indexed by node name

    '''
    translations = {}
    for sname, seq in aln.items():
        aa_seq = safe_translate(str(feature.extract(seq)))
        translations[sname] = aa_seq

    return translations


def translate_vcf_feature(sequences, ref, feature, feature_name):
    '''Translates a subsequence of input nucleotide sequences.

    Parameters
    ----------
    sequences : dict
        TreeTime format dictionary from VCF-input of sequences indexed by node name

    ref :
        reference alignment the VCF was mapped to

    feature : Bio.Seq.Seq
        BioPython sequence feature

    Returns
    -------
    dict :
        translated reference gene, positions of AA differences, and AA
        differences indexed by node name

    Raises
    ------
    NoVariationError : if no variable sites within this feature (across all sequences)
    '''
    def str_reverse_comp(str_seq):
        #gets reverse-compliment of a string and returns it as a string
        seq_str = Seq(str_seq)
        return str(seq_str.reverse_complement())

    from Bio.Seq import Seq
    from Bio.SeqRecord import SeqRecord
    prot = {}
    prot['sequences'] = {}
    prot['positions'] = []

    refNuc = str(feature.extract( SeqRecord(seq=Seq(ref)) ).seq)
    # Need to get ref translation to store. check if multiple of 3 for sanity.
    # will be padded in safe_translate if not
    if len(refNuc)%3:
        print(f"Gene length of {feature_name!r} is not a multiple of 3. will pad with N", file=sys.stderr)

    ref_aa_seq = safe_translate(refNuc)
    prot['reference'] = ref_aa_seq

    start = int(feature.location.start) #1 less than GFF file; Bio 'pythons' the numbering
    end = int(feature.location.end)

    for seqk in sequences.keys():
        varSite = np.array(list(sequences[seqk].keys()))  #get positions where nuc diffs
        #reduce to only those within current gene
        geneVarSites = np.logical_and(varSite >= start, varSite < end)
        nucVarSites = varSite[geneVarSites] #translate this back to nuc position
        genNucSites = nucVarSites-start #get it in position within the gene to ensure in frame

        #Translate just the codon this nuc diff is in, and find out which AA loc
        #But need numbering to be w/in protin, not whole genome
        if feature.location.strand == -1:
            aaRepLocs = {(end-start-i-1)//3:safe_translate( str_reverse_comp( "".join([sequences[seqk][key+start]
                                    if key+start in sequences[seqk].keys() else ref[key+start]
                                for key in range(i-i%3,i+3-i%3)]) ))
                            for i in genNucSites}
        else:
            aaRepLocs = {i//3:safe_translate( "".join([sequences[seqk][key+start]
                                if key+start in sequences[seqk].keys() else ref[key+start]
                            for key in range(i-i%3,i+3-i%3)]) )
                        for i in genNucSites}

        aaRepLocsFinal = {}
        #remove if is a synonymous mutation!
        for key,val in aaRepLocs.items():
            if ref_aa_seq[key] != val:
                aaRepLocsFinal[key] = val
        aaRepLocs = aaRepLocsFinal

        #store the dict of differences
        prot['sequences'][seqk] = aaRepLocs
        #add to list of positions if needed
        for key in aaRepLocs.keys():
            if key not in prot['positions']:
                prot['positions'].append(key)

    prot['positions'].sort()

    # raise an error if no variable sites observed
    if len(prot['positions']) == 0:
        raise NoVariationError()
    return prot

def construct_mut(start, pos, end):
    return str(start) + str(pos) + str(end)

def assign_aa_vcf(tree, translations):
    aa_muts = {}

    #get mutations on the root
    root = tree.root
    aa_muts[root.name]={"aa_muts":{}}
    #If has no root node name, exit with error
    if root.name is None:
        print("\n*** Can't find node name for the tree root!")
        raise MissingNodeError()

    for fname, prot in translations.items():
        if root.name not in prot['sequences']:
            print("\n*** Can't find %s in the alignment provided!"%(root.name))
            raise MismatchNodeError()
        root_muts = prot['sequences'][root.name]
        tmp = []
        for pos in prot['positions']:
            if pos in root_muts:
                tmp.append(construct_mut(prot['reference'][pos], int(pos+1), root_muts[pos]))
        aa_muts[root.name]["aa_muts"][fname] = tmp

    for n in tree.get_nonterminals():
        for c in n:
            aa_muts[c.name]={"aa_muts":{}}
        for fname, prot in translations.items():
            if n.name not in prot['sequences']:
                print("\n*** Can't find %s in the alignment provided!"%(root.name))
                raise MismatchNodeError()
            n_muts = prot['sequences'][n.name]
            for c in n:
                tmp = []
                if c.name is None:
                    print("\n*** Internal node missing a node name!")
                    raise MissingNodeError()
                c_muts = prot['sequences'][c.name]
                for pos in prot['positions']:
                    #if pos in both, check if same
                    if pos in n_muts and pos in c_muts:
                        if n_muts[pos] != c_muts[pos]:
                            tmp.append(construct_mut(n_muts[pos], int(pos+1), c_muts[pos]))
                    elif pos in n_muts:
                        tmp.append(construct_mut(n_muts[pos], int(pos+1), prot['reference'][pos]))
                    elif pos in c_muts:
                        tmp.append(construct_mut(prot['reference'][pos], int(pos+1), c_muts[pos]))

                aa_muts[c.name]["aa_muts"][fname] = tmp

    aa_muts[root.name]['aa_sequences'] = {}
    for gene_name, gene_data in translations.items():
        root_seq = list(gene_data['reference'])
        for pos,alt in gene_data['sequences'][root.name].items():
            # pos is 0-based, <class 'numpy.int64'>
            root_seq[pos] = alt
        aa_muts[root.name]['aa_sequences'][gene_name] = "".join(root_seq)

    return aa_muts

def assign_aa_fasta(tree, translations, reference_translations):
    aa_muts = {}

    # Depending on how `augur ancestral` was run, the input JSON (nt_muts.json)
    # may or may not have mutations defined on the root node. Note that the 'muts'
    # array will always be present, but it can only contain mutations if a
    # --root-sequence was provided to augur ancestral.

    root = tree.get_nonterminals()[0]

    for n in tree.get_nonterminals():
        if n.name is None:
            print("\n*** Tree is missing node names!")
            raise MissingNodeError()
        for c in n:
            aa_muts[c.name]={"aa_muts":{}}
        for fname, aln in translations.items():
            for c in n:
                if c.name in aln and n.name in aln:
                    tmp = [construct_mut(a, int(pos+1), d) for pos, (a,d) in
                            enumerate(zip(aln[n.name], aln[c.name])) if a!=d]
                    aa_muts[c.name]["aa_muts"][fname] = tmp
                elif c.name not in aln and n.name not in aln:
                    print("\n*** Can't find %s OR %s in the alignment provided!"%(c.name, n.name))
                    raise MismatchNodeError()
                else:
                    print("no sequence pair for nodes %s-%s"%(c.name, n.name))

        if n==tree.root:
            # The aa_sequences at the root are simply the translation from the root-node input data
            aa_muts[n.name]={"aa_sequences":{}}
            for fname, aln in translations.items():
                if n.name in aln:
                    aa_muts[n.name]["aa_sequences"][fname] = "".join(aln[n.name])
            # The aa_muts are found by comparing the aa_sequence with the reference sequence
            aa_muts[n.name]['aa_muts'] = {}
            for fname, aln in translations.items():
                ref = reference_translations[fname]
                muts = [construct_mut(a, int(pos+1), d) for pos, (a,d) in enumerate(zip(ref, aln[n.name])) if a!=d]
                aa_muts[n.name]["aa_muts"][fname] = muts

    return aa_muts

def sequences_vcf(reference_fasta, vcf):
    """
    Extract the nucleotide variation in the VCF
    Returns a tuple
    [0] The sequences as a dict of dicts. sequences → <NODE_NAME> → <POS> → <ALT_NUC> where <POS> is a 0-based int
    [1] The sequence of the provided `reference_fasta` (string)
    """
    assert reference_fasta is not None
    compress_seq = read_vcf(vcf, reference_fasta)
    sequences = compress_seq['sequences']
    ref = compress_seq['reference']
    return (sequences, ref)

def sequences_json(node_data_json, tree, validation_mode):
    """
    Extract the full nuc sequence for each node in the provided node-data JSON.
    Returns a dict, keys are node names and values are a string of the genome sequence (nuc)
    """
    node_data = read_node_data(node_data_json, validation_mode=validation_mode)
    if node_data is None:
        raise AugurError("could not read node data (incl sequences)")
    # extract sequences from node meta data
    sequences = {}
    for k,v in node_data['nodes'].items():
        if 'sequence' in v:
            sequences[k] = v['sequence']
    tree_nodes = {c.name for c in tree.find_clades()}
    tree_nodes_missing_sequences = tree_nodes - set(sequences.keys())
    if len(tree_nodes_missing_sequences):
        raise AugurError(dedent(f"""\
            {len(tree_nodes_missing_sequences)} nodes on the tree are missing nucleotide sequences in the node-data JSON.
            These must be present under 'nodes' → <node_name> → 'sequence'.
            This error may originate from using 'augur ancestral' with VCF input; in this case try using VCF output from that command here.
            """))
    reference = node_data['reference']['nuc']
    return (reference, sequences)

def register_parser(parent_subparsers):
    parser = parent_subparsers.add_parser("translate", help=__doc__)
    parser.add_argument('--tree', required=True, help="prebuilt Newick -- no tree will be built if provided")
    parser.add_argument('--ancestral-sequences', required=True, type=str, help='JSON (fasta input) or VCF (VCF input) containing ancestral and tip sequences')
    parser.add_argument('--reference-sequence', required=True,
                        help='GenBank or GFF file containing the annotation')
    parser.add_argument('--genes', nargs='+', action='extend', help="genes to translate (list or file containing list)")
    parser.add_argument('--output-node-data', type=str, help='name of JSON file to save aa-mutations to')
    parser.add_argument('--alignment-output', type=str, help="write out translated gene alignments. "
                                   "If a VCF-input, a .vcf or .vcf.gz will be output here (depending on file ending). If fasta-input, specify the file name "
                                   "like so: 'my_alignment_%%GENE.fasta', where '%%GENE' will be replaced by the name of the gene")
    add_validation_arguments(parser)

    vcf_only = parser.add_argument_group(
        title="VCF specific",
        description="These arguments are only applicable if the input (--ancestral-sequences) is in VCF format."
    )
    vcf_only.add_argument('--vcf-reference', type=str, help='fasta file of the sequence the VCF was mapped to')
    vcf_only.add_argument('--vcf-reference-output', type=str, help="fasta file where reference sequence translations for VCF input will be written")

    return parser

def check_arg_combinations(args, is_vcf):
    """
    Check that provided arguments are compatible.
    Where possible we use argparse built-ins, but they don't cover everything we want to check.
    This checking shouldn't be used by downstream code to assume arguments exist, however by checking for
    invalid combinations up-front we can exit quickly.
    """

    if is_vcf:
        if not args.vcf_reference:
            raise AugurError("A reference FASTA (--vcf-reference) is required with VCF-format input")
    else:
        if args.vcf_reference or args.vcf_reference_output:
            raise AugurError("Arguments '--vcf-reference' and/or '--vcf-reference-output' are only applicable if the input ('--ancestral-sequences') is VCF")
    
    if args.alignment_output:
        if is_vcf:
            if not is_filename_vcf(args.alignment_output):
                raise AugurError("When using a VCF input the --alignment-output filename must also be a VCF file")
            if not args.vcf_reference_output:
                raise AugurError("When using a VCF input and --alignment-output, we now require you to specify the --vcf-reference-output as well")
        else:
            if is_filename_vcf(args.alignment_output):
                raise AugurError("When using a non-VCF input the --alignment-output filename must not be a VCF file")
    if args.vcf_reference_output and not args.alignment_output:
        raise AugurError("The VCF reference output (--vcf-reference-output) needs --alignment-output")


def run(args):
    ## read tree and data, if reading data fails, return with error code
    tree = Phylo.read(args.tree, 'newick')
    is_vcf = is_filename_vcf(args.ancestral_sequences)
    check_arg_combinations(args, is_vcf)

    genes = parse_genes_argument(args.genes)

    ## load features; only requested features if genes given
    features = load_features(args.reference_sequence, genes)
    print("Read in {} features from reference sequence file".format(len(features)))

    ## Read in sequences & for each sequence translate each feature _except for_ the 'nuc' feature name
    ## Note that except for the 'nuc' annotation, `load_features` _only_ looks for 'gene' (GFF files) or 'CDS' (GenBank files)
    ## The reference translations are straight translations of the "reference.nuc" sequence in the JSON
    ## (see <https://github.com/nextstrain/augur/issues/1362> for more details here), _or_ for VCF input a translation
    ## from the provided FASTA reference
    translations = {}
    reference_translations = {}
    if is_vcf:
        (sequences, ref) = sequences_vcf(args.vcf_reference, args.ancestral_sequences)
        features_without_variation = []
        for fname, feat in features.items():
            if fname=='nuc':
                continue
            try:
                translations[fname] = translate_vcf_feature(sequences, ref, feat, fname)
                reference_translations[fname] = translations[fname]['reference']
            except NoVariationError:
                features_without_variation.append(fname)
        if len(features_without_variation):
            print("{} genes had no mutations and so have been be excluded.".format(len(features_without_variation)))  
    else:
        (reference, sequences) = sequences_json(args.ancestral_sequences, tree, args.validation_mode)
        translations = {fname: translate_feature(sequences, feat) for fname, feat in features.items() if fname!='nuc'}
        for fname, feat in features.items():
            if fname=='nuc':
                continue
            reference_translations[fname] = safe_translate(str(feat.extract(reference)))

    annotations = genome_features_to_auspice_annotation(features, args.reference_sequence, assert_nuc=True)

    ## determine amino acid mutations for each node
    try:
        if is_vcf:
            aa_muts = assign_aa_vcf(tree, translations)
        else:
            aa_muts = assign_aa_fasta(tree, translations, reference_translations)
    except MissingNodeError as err:
        print("\n*** ERROR: Some/all nodes have no node names!")
        print("*** Please check you are providing the tree output by 'augur refine'.")
        print("*** If you haven't run 'augur refine', please add node names to your tree by running:")
        print("*** augur refine --tree %s --output-tree <filename>.nwk"%(args.tree) )
        print("*** And use <filename>.nwk as the tree when running 'ancestral', 'translate', and 'traits'")
        return 1
    except MismatchNodeError as err:
        print("\n*** ERROR: Mismatch between node names in %s and in %s"%(args.tree, args.ancestral_sequences))
        print("*** Ensure you are using the same tree you used to run 'ancestral' as input here.")
        print("*** Or, re-run 'ancestral' using %s, then use the new %s as input here."%(args.tree, args.ancestral_sequences))
        return 1

    output_data = {'annotations':annotations, 'nodes':aa_muts, 'reference': reference_translations}
    out_name = get_json_name(args, '.'.join(args.tree.split('.')[:-1]) + '_aa-mutations.json')
    # use NodeDataObject to perform validation on the file before it's written
    NodeDataObject(output_data, out_name, args.validation_mode)

    write_json(output_data, out_name)
    print("amino acid mutations written to", out_name, file=sys.stdout)

    ## write alignments to file if requested
    if args.alignment_output:
        if is_vcf:
            assert is_filename_vcf(args.alignment_output)
            assert args.vcf_reference_output is not None
            write_VCF_translation(translations, args.alignment_output, args.vcf_reference_output)
        else:
            ## write fasta-style output if requested
            if '%GENE' in args.alignment_output:
                for fname, seqs in translations.items():
                    SeqIO.write([SeqRecord.SeqRecord(seq=Seq.Seq(s), id=sname, name=sname, description='')
                                 for sname, s in seqs.items()],
                                 args.alignment_output.replace('%GENE', fname), 'fasta')
            else:
                print("ERROR: alignment output file does not contain '%GENE', so will not be written.")



r"""
Write a file like Augur, with transparent optimized compression and universal newlines.

Input is always from stdin.

Output is to the given file path, as the compression format detection require
it.  The given path may be "-" to explicitly write to stdout, but no
decompression will be done.

Universal newline translation is always performed, so \n, \r\n, and \r in the
input are all translated to the system's native newlines (e.g. \n on Unix, \r\n
on Windows) in the output.
"""
import io
import sys
from shutil import copyfileobj

from .io.file import open_file
from .utils import first_line


# The buffer size used by xopen() (which underlies open_file()), which notes:
# 128KB [KiB] buffer size also used by cat, pigz etc. It is faster than the 8K
# [KiB] default.
BUFFER_SIZE = max(io.DEFAULT_BUFFER_SIZE, 128 * 1024)


def register_parser(parent_subparsers):
    parser = parent_subparsers.add_parser("write-file", help=first_line(__doc__))
    parser.add_argument("path", metavar="PATH", help="path to file")
    return parser


def run(args):
    with open_file(args.path, "wt", newline=None) as f:
        # It's tempting to want to splice(2) here, but it turns out to make
        # little sense.  Firstly, the availability of splice(2) is Linux,
        # Python ≥3.10, and one of the files needs to be a pipe.  The chance of
        # all of those together is slim-to-none, particularly because even in
        # the common case of xopen() reading from a pipe—the stdout of an
        # external decompression process—we can't use that pipe directly
        # because xopen() always buffers the first block of the file into
        # Python.¹  Secondly, we want universal newline handling—so that
        # callers get behaviour consistent with the rest of Augur—and that
        # rules out splice(2).
        #
        # Copying the data thru Python instead of with splice(2) seems fast
        # enough in some quick trials with large files (e.g. against `zstd`
        # directly), and the bottlenecks in pipelines using this command will
        # often not be this command's i/o.
        #   -trs, 11 July 2024, updated 24 July 2024
        #
        # ¹ <https://github.com/pycompression/xopen/blob/67651844/src/xopen/__init__.py#L374>
        copyfileobj(sys.stdin, f, BUFFER_SIZE)



"""
Verifies strain name pattern in the 'strain' field of the NDJSON
record. Adds a 'strain' field to the record if it does not already
exist.
"""

import argparse
import re
from typing import Generator, List
from augur.io.print import print_err
from augur.utils import first_line


def transform_name(
    record: dict,
    index: int,
    strain_name_pattern: re.Pattern,
    backup_fields: List[str],
) -> dict:
    # Verify strain name matches the strain regex pattern
    if strain_name_pattern.match(record.get("strain", "")) is None:
        # Default to empty string if not matching pattern
        record["strain"] = ""

        # Use non-empty value of backup fields if provided
        if backup_fields:
            for field in backup_fields:
                if record.get(field):
                    record["strain"] = str(record[field])
                    break

    if record["strain"] == "":
        print_err(
            f"WARNING: Record number {index} has an empty string as the strain name.",
        )

    return record


def register_parser(
    parent_subparsers: argparse._SubParsersAction,
) -> argparse._SubParsersAction:
    parser = parent_subparsers.add_parser(
        "transform-strain-name",
        parents=[parent_subparsers.shared_parser],  # type: ignore
        help=first_line(__doc__),
    )

    parser.add_argument(
        "--strain-regex",
        default="^.+$",
        help="Regex pattern for strain names. "
        + "Strain names that do not match the pattern will be dropped.",
    )
    parser.add_argument(
        "--backup-fields",
        nargs="*",
        default=[],
        help="List of backup fields to use as strain name if the value in 'strain' "
        + "does not match the strain regex pattern. "
        + "If multiple fields are provided, will use the first field that has a non-empty string.",
    )

    return parser


def run(args: argparse.Namespace, records: List[dict]) -> Generator[dict, None, None]:
    strain_name_pattern = re.compile(args.strain_regex)

    for index, record in enumerate(records):
        transform_name(
            record,
            index,
            strain_name_pattern,
            args.backup_fields,
        )

        yield record



from pathlib import Path
from subprocess import run

topdir = Path(__file__).resolve().parent.parent

def test_flake8():
    # Check the exit status ourselves for nicer test output on failure
    result = run(["flake8"], cwd = topdir)
    assert result.returncode == 0, "flake8 exited with errors"
